/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      FileDialog
Version:   1.0
ID:        ID_FILEDIALOG
Status:    Beta
Category:  GUI
Date:      January 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2003-2004.  All rights reserved.
Keywords:  dialog, box, file, directory, query, user, request, requester
Short:     The FileDialog class is used to request file selections from the user.

###DESCRIPTION###
<p>The FileDialog class provides a standard method for prompting the user
with a file requester.  The process of initialisation is straight forward
and the class is designed both for programs and script usage.</p>

<p>The following example illustrates how you can create a file requester that
is appropriate for selecting text files.  Notice our use of the Set statement,
which will write the user's selected file to an external object.</p>

<pre>
  &lt;filedialog target="[desktop]" title="Select a Text File"
    filterlist="All Files;*;Text Files;*.txt;DML Files;*.dml"
    location="documents:"&gt;
    &lt;set static object="[myobject]" &string="\[owner.location\]"/&gt;
  &lt;/filedialog&gt;
</pre>

<p>The dialog box may be configured in a number of areas, including filter
definitions, button text and the window title.  Please refer to the
documentation for the file dialog fields for more information.</p>

<p>If the dialog box needs to be used multiple times, create it as static
and then use the Show action to display the dialog window as required.  This
is a good way of 'caching' the window so that it does not need to be recreated
from scratch each time that the dialog window needs to be displayed.</p>

<p>Any child objects that are initialised to a dialog will be activated in the
event that a successful response is given by the user.  Failure to respond,
or some form of cancellation on the user's part will prevent the activation
of the child objects.</p>

<p>If you are programming the file dialog directly and would like to receive
notification of success, subscribe to the Activate action.</p>

###END###

*****************************************************************************/

//#define DEBUG

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/data/xml.h>
#include <pandora/tools/dialog.h>
#include <pandora/tools/filedialog.h>
#include <pandora/tools/fileview.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_FILEDIALOG, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2003-2004. All rights reserved.",
   "January 2004",
   "FileDialog"
};

static struct KernelBase  *KernelBase = NULL;
static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static OBJECTPTR FileDialogClass = NULL;
static OBJECTPTR StringsMod = NULL;
static OBJECTPTR FileSystemMod = NULL;
static LONG glBreakMessageID = 0;

#define BUTTONGAP   6

/*****************************************************************************
** Object structure.
*/

struct FileDialog {
   OBJECT_HEADER             /* Class header */
   OBJECTID WindowID;        /* Reference to the window of the dialog box */
   OBJECTID TargetID;        /* The target object (usually the desktop) */
   OBJECTID AutoSaveID;      /* Auto-Save from this object reference */
   LONG     Flags;           /* Special options */
   LONG     Response;        /* The resulting response from the user */
   LONG     Static;          /* TRUE if object is static */
   STRING   *SelectedFiles;  /* File selection array */

   PRIVATE_FIELDS
   UBYTE     Title[80];       /* Window title */
   UBYTE     Icon[100];       /* Window icon location */
   UBYTE     Location[300];
   UBYTE     Filter[60];      /* Current filter text (expressed as a wildcard and not a filter name) */
   UBYTE     OkayText[40];
   UBYTE     CancelText[40];
   STRING    FilterList;
   OBJECTPTR ScriptObject;
   MEMORYID  ScriptMID;
   STRING    Script;          /* Simple DML statement indicating what to do on button-press, e.g. "<set object="[desktop]" field="">"... */
   OBJECTPTR LocationBox;     /* Reference to the input box's text object, for when input boxes are used */
   OBJECTPTR FilterBox;
   OBJECTPTR OkayButton;
   OBJECTPTR CancelButton;
   OBJECTPTR FileView;
   OBJECTPTR View;
   OBJECTPTR PathBox;
   WORD      Active;
};

/*****************************************************************************
** Class definition.
*/

static struct FieldDef FileDialogFlags[] = {
   { "WAIT",          FDF_WAIT },
   { "STRICTFILTER",  FDF_STRICTFILTER },
   { "MULTISELECT",   FDF_MULTISELECT },
   { "NOFILECHECK",   FDF_NOFILECHECK },
   { "CONFIRMEXISTS", FDF_CONFIRMEXISTS },
   { "SAVE",          FDF_SAVE },
   { "SELECTDIR",     FDF_SELECTDIR },
   { "NOFILES",       FDF_NOFILES },
   { NULL, NULL }
};

static struct FieldDef ResponseTable[] = {
   { "OKAY",   RSF_OKAY },
   { "CANCEL", RSF_CANCEL },
   { NULL, NULL }
};

static ERROR GET_ActionScript(struct FileDialog *, STRING *);
static ERROR GET_CancelText(struct FileDialog *, STRING *);
static ERROR GET_Filter(struct FileDialog *, STRING *);
static ERROR GET_FilterList(struct FileDialog *, STRING *);
static ERROR GET_Icon(struct FileDialog *, STRING *);
static ERROR GET_Location(struct FileDialog *, STRING *);
static ERROR GET_OkayText(struct FileDialog *, STRING *);
static ERROR GET_Title(struct FileDialog *, STRING *);

static ERROR SET_ActionScript(struct FileDialog *, STRING);
static ERROR SET_AutoSave(struct FileDialog *, OBJECTID *);
static ERROR SET_CancelText(struct FileDialog *, STRING);
static ERROR SET_Filter(struct FileDialog *, STRING);
static ERROR SET_FilterList(struct FileDialog *, STRING);
static ERROR SET_Icon(struct FileDialog *, STRING);
static ERROR SET_Location(struct FileDialog *, STRING);
static ERROR SET_OkayText(struct FileDialog *, STRING);
static ERROR SET_Title(struct FileDialog *, STRING);

static struct FieldArray FileDialogFields[] = {
   { "Window",         0, FDF_OBJECTID|FDF_RW,        0, NULL, NULL },
   { "Target",         0, FDF_OBJECTID|FDF_RI,        0, NULL, NULL },
   { "AutoSave",       0, FDF_OBJECTID|FDF_RW,        0, NULL, SET_AutoSave },
   { "Flags",          0, FDF_LONGFLAGS|FDF_RW,       (LONG)&FileDialogFlags, NULL, NULL },
   { "Response",       0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&ResponseTable, NULL, NULL },
   { "Static",         0, FDF_LONG|FDF_RW,            0, NULL, NULL },
   { "SelectedFiles",  0, FDF_STRING|FDF_R,           0, NULL, NULL },
   /** Virtual fields ***/
   { "ActionScript",   0, FDF_STRING|FDF_RW,    0, GET_ActionScript,  SET_ActionScript },
   { "CancelText",     0, FDF_STRING|FDF_RW,    0, GET_CancelText,    SET_CancelText },
   { "Filter",         0, FDF_STRING|FDF_RW,    0, GET_Filter,        SET_Filter },
   { "FilterList",     0, FDF_STRING|FDF_RW,    0, GET_FilterList,    SET_FilterList },
   { "Location",       0, FDF_STRING|FDF_RW,    0, GET_Location,      SET_Location },
   { "Icon",           0, FDF_STRING|FDF_RW,    0, GET_Icon,          SET_Icon },
   { "OkayText",       0, FDF_STRING|FDF_RW,    0, GET_OkayText,      SET_OkayText },
   { "Script",         0, FDF_STRING|FDF_RW,    0, GET_ActionScript,  SET_ActionScript },
   { "Title",          0, FDF_STRING|FDF_RW,    0, GET_Title,         SET_Title },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR FILEDIALOG_ActionNotify(struct FileDialog *, struct acActionNotify *);
static ERROR FILEDIALOG_Activate(struct FileDialog *, APTR);
static ERROR FILEDIALOG_ClosingTag(struct FileDialog *, APTR);
static ERROR FILEDIALOG_FieldNotify(struct FileDialog *, struct acFieldNotify *);
static ERROR FILEDIALOG_Free(struct FileDialog *, APTR);
static ERROR FILEDIALOG_Init(struct FileDialog *, APTR);
static ERROR FILEDIALOG_NewObject(struct FileDialog *, APTR);
static ERROR FILEDIALOG_SetUnlistedField(struct FileDialog *, struct acSetUnlistedField *);
static ERROR FILEDIALOG_Show(struct FileDialog *, APTR);

static struct ActionArray FileDialogActions[] = {
   { AC_ActionNotify,     FILEDIALOG_ActionNotify },
   { AC_Activate,         FILEDIALOG_Activate },
   { AC_ClosingTag,       FILEDIALOG_ClosingTag },
   { AC_FieldNotify,      FILEDIALOG_FieldNotify },
   { AC_Free,             FILEDIALOG_Free },
   { AC_Init,             FILEDIALOG_Init },
   { AC_NewObject,        FILEDIALOG_NewObject },
   { AC_SetUnlistedField, FILEDIALOG_SetUnlistedField },
   { AC_Show,             FILEDIALOG_Show },
   { NULL, NULL }
};

/*****************************************************************************
** Prototypes
*/

static ERROR CreateWindow(struct FileDialog *);
static LONG OpenFile(struct FileDialog *);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &FileSystemMod, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(FileSystemMod, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   glBreakMessageID = AllocateID(IDTYPE_MESSAGE);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&FileDialogClass, NULL,
      FID_BaseClassID|TLONG, ID_FILEDIALOG,
      FID_Version|TFLOAT,    VER_FILEDIALOG,
      FID_Name|TSTRING,      "FileDialog",
      FID_Category|TLONG,    CCF_TOOL,
      FID_Flags|TLONG,       CLF_PROMOTECHILDREN|CLF_PRIVATEONLY,
      FID_Actions|TPTR,      FileDialogActions,
      FID_Fields|TPTR,       FileDialogFields,
      FID_Size|TLONG,        sizeof(struct FileDialog),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (FileDialogClass) { Action(AC_Free, FileDialogClass, NULL); FileDialogClass = NULL; }
   if (StringsMod)      { Action(AC_Free, StringsMod, NULL); StringsMod = NULL; }
   if (FileSystemMod)   { Action(AC_Free, FileSystemMod, NULL); FileSystemMod = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** FileDialog: ActionNotify
*/

static ERROR FILEDIALOG_ActionNotify(struct FileDialog *Self, struct acActionNotify *Args)
{
   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if ((Args->ActionID IS AC_Free) AND (Args->ObjectID IS Self->WindowID)) {
      if (Self->Active) SendMessage(NULL, glBreakMessageID, NULL, NULL, NULL);
      else {
         if (Self->WindowID) {
            acFreeID(Self->WindowID);
            Self->WindowID = NULL;
         }
         if (Self->Static IS FALSE) acFree(Self);
      }

      Self->LocationBox  = NULL;
      Self->FilterBox    = NULL;
      Self->OkayButton   = NULL;
      Self->CancelButton = NULL;
      Self->FileView     = NULL;
      Self->View         = NULL;
      Self->PathBox      = NULL;
   }
   else if (Args->ActionID IS AC_Activate) {
      /*** React to the object that has been activated ***/

      if (Args->ObjectID IS Self->PathBox->UniqueID) {
         STRING str;
         GetField(Self->PathBox, FID_String, FT_STRING, &str);
         SetField(Self->FileView, FID_Location, FT_STRING, str);
      }
      else if ((Args->ObjectID IS Self->OkayButton->UniqueID) OR
          (Args->ObjectID IS Self->FileView->UniqueID) OR
          (Args->ObjectID IS Self->LocationBox->UniqueID)) {

         if (OpenFile(Self) IS TRUE) {
            if (Self->WindowID) {
               ActionMsg(AC_Free, Self->WindowID, NULL);
               Self->WindowID = NULL;
            }
            if (Self->Static IS FALSE) Action(AC_Free, Self, NULL);
         }
      }
      else if (Args->ObjectID IS Self->CancelButton->UniqueID) {
         Self->Response = RSF_CANCEL;

         if (Self->WindowID) {
            ActionMsg(AC_Free, Self->WindowID, NULL);
            Self->WindowID = NULL;
         }

         if (Self->Static IS FALSE) Action(AC_Free, Self, NULL);
      }
      else if (Args->ObjectID IS Self->FilterBox->UniqueID) {
         STRING filter, str;
         UBYTE text[120], filtervalue[120];
         WORD i;

         /*** Write the filter to the fileview ***/

         DPrintF("~ActionNotify:","[FileDialog] Filter change detected.");

            GetField(Self->FilterBox, FID_String, FT_STRING, &str);

            /*** Check what has been entered against the known filter names ***/

            filter = Self->FilterList;
            while (*filter) {
               for (i=0; (*filter) AND (*filter != ';') AND (i < sizeof(text)-1); i++) text[i] = *filter++;
               text[i] = 0;
               if (*filter IS ';') filter++;

               for (i=0; (*filter) AND (*filter != ';') AND (i < sizeof(filtervalue)-1); i++) filtervalue[i] = *filter++;
               filtervalue[i] = 0;
               if (*filter IS ';') filter++;

               if (StrCompare(str, text, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                  str = filtervalue;
                  break;
               }
            }

            /*** Set the filter ***/

            SetField(Self->FileView, FID_Filter, FT_STRING, str);

         StepBack();
      }
      else DPrintF("@ActionNotify:","[FileDialog] Activate action on unuspported object #%d.", Args->ObjectID);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** FileDialog: Activate
*/

static ERROR FILEDIALOG_Activate(struct FileDialog *Self, APTR Void)
{
   /* Note: See the OpenFile() function for Activate notification support.
   ** If the user calls Activate(), it simply presents the file dialog
   ** (useful for static FileDialog objects).
   */

   return(acShow(Self)|ERF_Notified);
}

/*****************************************************************************
** FileDialog: ClosingTag
*/

static ERROR FILEDIALOG_ClosingTag(struct FileDialog *Self, APTR Void)
{
   if (Self->Static IS FALSE) {
      DPrintF("~ClosingTag()","[Dialog]");
         acShow(Self);
         /* Note: The dialog window will be destroyed automatically
         ** when a response is received from the buttons.
         */
      StepBack();
   }
   return(ERR_Okay);
}

/*****************************************************************************
** FileDialog: FieldNotify
*/

static ERROR FILEDIALOG_FieldNotify(struct FileDialog *Self, struct acFieldNotify *Args)
{
   UBYTE buffer[400];
   STRING selection;
   LONG type;

   if (!Args) return(ObjectError(Self, ERH_FieldNotify, ERR_Args));

   if (Args->FieldID IS FID_SelectedTag) {
      /*** The user has selected an item in the file view ***/

      if (!Self->FileView) return(ERR_Okay);

      if ((GetField(Self->FileView, FID_SelectionPath, FT_STRING, &selection) != ERR_Okay) OR (!selection)) {
         if (Self->LocationBox) SetField(Self->LocationBox, FID_String, FT_STRING, "");
         return(ERR_Okay);
      }

      if (AnalyseLocation(selection, &type) != ERR_Okay) {
         if (Self->LocationBox) SetField(Self->LocationBox, FID_String, FT_STRING, "");
         return(ERR_Okay);
      }

      if ((type IS LOC_FILE) OR
          ((Self->Flags & FDF_SELECTDIR) AND ((type IS LOC_DIRECTORY) OR (type IS LOC_ASSIGN)))) {
         GetField(Self->FileView, FID_Selection, FT_STRING, &selection);

         if (type IS LOC_DIRECTORY) {
            StrFormat(buffer, sizeof(buffer), "%s/", selection);
            selection = buffer;
         }
         else if (type IS LOC_ASSIGN) {
            StrFormat(buffer, sizeof(buffer), "%s:", selection);
            selection = buffer;
         }

         SetField(Self->LocationBox, FID_String, FT_STRING, selection);
      }
      else SetField(Self->LocationBox, FID_String, FT_STRING, "");
   }
   else if (Args->FieldID IS FID_Location) {
      STRING path;
      if (GetField(Self->FileView, FID_Location, FT_STRING, &path) IS ERR_Okay) {
         SetField(Self->PathBox, FID_String, FT_STRING, path);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** FileDialog: Free
*/

static ERROR FILEDIALOG_Free(struct FileDialog *Self, APTR Void)
{
   OBJECTPTR window;

   if (Self->SelectedFiles) {
      FreeMemory(Self->SelectedFiles);
      Self->SelectedFiles = NULL;
   }

   if (Self->ScriptObject) {
      acFree(Self->ScriptObject);
      Self->ScriptObject = NULL;
   }

   if (Self->ScriptMID) {
      if (Self->Script) {
         ReleaseMemory(Self->Script);
         Self->Script = NULL;
      }
      FreeMemoryID(Self->ScriptMID);
      Self->ScriptMID = NULL;
   }

   if ((Self->WindowID) AND (CheckObjectExists(Self->WindowID, NULL) IS ERR_Okay)) {
      if (AccessObject(Self->WindowID, 5000, &window) IS ERR_Okay) {
         UnsubscribeAction(window, NULL, Self->Head.UniqueID);
         ReleaseObject(window);
      }

      acFreeID(Self->WindowID);
      Self->WindowID = NULL;
   }

   if (Self->FilterList) {
      FreeMemory(Self->FilterList);
      Self->FilterList = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** FileDialog: Init()
*/

static ERROR FILEDIALOG_Init(struct FileDialog *Self, APTR Void)
{
   ERROR error;

   /* Create the dialog window now if we are not static, otherwise
   ** we can wait until the Show action is used.
   */

   if (Self->Static IS FALSE) {
      #ifndef DEBUG
      DebugState(FALSE, -1);
      #endif

      error = CreateWindow(Self);

      #ifndef DEBUG
      DebugState(TRUE, -1);
      #endif

      if (error != ERR_Okay) return(ObjectError(Self, ERH_Init, error));
   }

   return(ERR_Okay);
}

/*****************************************************************************
** FileDialog: NewObject()
*/

static ERROR FILEDIALOG_NewObject(struct FileDialog *Self, APTR Void)
{
   StrCopy("Please Select a File", Self->Title, sizeof(Self->Title));
   StrCopy("icons:folders/folder", Self->Icon, sizeof(Self->Icon));
   StrCopy("Open File", Self->OkayText, sizeof(Self->OkayText));
   StrCopy("Cancel", Self->CancelText, sizeof(Self->CancelText));
   SET_FilterList(Self, "All Files;*");
   Self->Response = RSF_CANCEL;
   return(ERR_Okay);
}

/*****************************************************************************
** FileDialog: SetUnlistedField()
*/

static ERROR FILEDIALOG_SetUnlistedField(struct FileDialog *Self, struct acSetUnlistedField *Args)
{
   if (!Args) return(ObjectError(Self, ERH_SetUnlistedField, ERR_Args));

   if (!Self->ScriptObject) {
      if (GET_ActionScript(Self, &Self->Script) IS ERR_Okay) {
         CreateObject(ID_SCRIPT, NF_CHILD, &Self->ScriptObject, NULL,
            FID_Location|TSTRING, Self->Script,
            TAGEND);
      }
   }

   if (!Self->ScriptObject) return(ObjectError(Self, ERH_SetUnlistedField, ERR_Failed));

   return(Action(AC_SetUnlistedField, Self->ScriptObject, Args));
}

/*****************************************************************************

###ACTION###
Name: Show
Short: Puts the dialog window on display.
###END###

*****************************************************************************/

static ERROR FILEDIALOG_Show(struct FileDialog *Self, APTR Void)
{
   struct Message message;
   struct NextMsg nextmsg;

   DPrintF("~Show()","[FileDialog:%d]", Self->Head.UniqueID);

   /*** If we are active, do not continue ***/

   if (Self->Active IS TRUE) return(ERR_Okay);
   Self->Active = TRUE;

   Self->Response = NULL;

   /* If our dialog window has disappeared (e.g. the user killed it on
   ** a previous activation), we'll need to recreate it.
   */

   if ((!Self->WindowID) OR (CheckObjectExists(Self->WindowID, NULL) != ERR_Okay)) {
      Self->WindowID = NULL;
      if (CreateWindow(Self) != ERR_Okay) return(ERR_Failed);
   }

   /*** Show the dialog window ***/

   ActionMsg(AC_MoveToFront, Self->WindowID, NULL);
   ActionMsg(AC_Show, Self->WindowID, NULL);
   ActionMsg(AC_Focus, Self->OkayButton->UniqueID, NULL);

   if (Self->Flags & DF_WAIT) {

      /* Wait for a user response.  We will awaken if the Response field
      ** is updated, or if the dialog window is killed.  See the code for
      ** the Response field for further details.
      */

      DPrintF("Show:","[Dialog] Entering sleep mode...");

      while (ProcessMessages(&nextmsg, MF_WAIT) IS ERR_Okay) {
         if (nextmsg.Count > 0) {
            if (GetMessage(NULL, NULL, NULL, &message, sizeof(struct Message)) IS ERR_Okay) {
               if ((message.Type IS MSGID_QUIT) OR (message.Type IS glBreakMessageID)) break;
               else DPrintF("@FileDialog:","Received unknown message type #%d.", message.Type);
            }
         }
      }

      DPrintF("Show:","[Dialog] Break received - I no longer slumber!");

      /*** Terminate our dialog object if we are non-static ***/

      if (Self->Static IS FALSE) Action(AC_Free, Self, NULL);
   }

   Self->Active = FALSE;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ActionScript
Short:  Script to be executed when the field dialog is activated.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>When a file dialog box is activated due to the user's confirmed selection
of a file, you can elect to run a script that performs a responsive action.
The ActionScript string will typically refer to a file location, although
the special STRING: assignment followed with script execution text (e.g.
DML) can also be useful for optimising execution for simple scripts.</p>

###END###

*****************************************************************************/

static ERROR GET_ActionScript(struct FileDialog *Self, STRING *Value)
{
   if (Self->Script) {
      *Value = Self->Script;
      return(ERR_Okay);
   }
   else if (!Self->ScriptMID) {
      *Value = NULL;
      return(ERR_FieldNotSet);
   }
   else if (AccessMemory(Self->ScriptMID, MEM_READ, (void **)&Self->Script) IS ERR_Okay) {
      *Value = Self->Script;
      return(ERR_Okay);
   }
   else {
      *Value = NULL;
      return(ObjectError((OBJECTPTR)Self, ERH_GetField, ERR_AccessMemory));
   }
}

static ERROR SET_ActionScript(struct FileDialog *Self, STRING Value)
{
   LONG i;

   if (Self->Script)    { ReleaseMemory(Self->Script);   Self->Script = NULL; }
   if (Self->ScriptMID) { FreeMemoryID(Self->ScriptMID); Self->ScriptMID = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR|Self->Head.MemFlags, (void **)&Self->Script, &Self->ScriptMID) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->Script[i] = Value[i];
         Self->Script[i] = 0;
      }
      else return(ObjectError((OBJECTPTR)Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   AutoSave
Short:  Simplifies the saving of data from objects to files.
Type:   OBJECTID
Status: Read/Set

###DESCRIPTION###


###END###

*****************************************************************************/

static ERROR SET_AutoSave(struct FileDialog *Self, OBJECTID *Value)
{
   if ((Self->AutoSaveID = *Value)) Self->Flags |= FDF_SAVE;
   else Self->Flags &= ~FDF_SAVE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   CancelText
Short:  Defines the string to use for the cancel button.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The text that appears in the cancel button of the file dialog may be
redefined here.  By default, the dialog will displaye the word "Cancel" in
the button.</p>

###END###

*****************************************************************************/

static ERROR GET_CancelText(struct FileDialog *Self, STRING *Value)
{
   if (Self->CancelText[0]) {
      *Value = Self->CancelText;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_CancelText(struct FileDialog *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->CancelText)-1) AND (Value[i]); i++) Self->CancelText[i] = Value[i];
      Self->CancelText[i] = 0;
   }
   else Self->CancelText[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Filter
Short:  The current filter string in the file dialog.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The file filter that is used for the dialog box can be defined by setting
this field.  By default the Filter is set to the first filter identified
from the FilterList field.  If that is not available then the filter string
will be empty and all files will be shown.</p>

<p>The filter that you set must either be a standard wildcard filter (such
as *.txt|*.text), or the name of a predefined filter from the FilterList
string.</p>

###SEE ALSO###
Field: FilterList

###END###

*****************************************************************************/

static ERROR GET_Filter(struct FileDialog *Self, STRING *Value)
{
   if (Self->Filter[0]) {
      *Value = Self->Filter;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Filter(struct FileDialog *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->Filter)-1) AND (Value[i]); i++) Self->Filter[i] = Value[i];
      Self->Filter[i] = 0;
   }
   else Self->Filter[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   FilterList
Short:  Defines a list of filter options to be displayed in the filter combo box.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>Suitable filters for the filter combo box are defined by setting this
field. Filters are listed in a single string, using the format
"text;value;text;value;...".  The "text" string is the name of the filter as
you want to display to the user, and the "value" string is the actual file
filter that will be used for filtering the files.  The following examples
illustrate:</p>

<pre>
   All Files;*;Text Files;*.txt|*.text;DML Files;*.dml
   PCX Pictures;*.pcx;PNG Pictures;*.png;JPEG Pictures;*.jpeg|*.jpg
</pre>

<p>The first filter that is defined in the list is used as the default.  If
you need to set a different default filter, you can do so by setting the
Filter field.</p>

###SEE ALSO###
Field: Filter

###END###

*****************************************************************************/

static ERROR GET_FilterList(struct FileDialog *Self, STRING *Value)
{
   if (Self->FilterList) {
      *Value = Self->FilterList;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_FilterList(struct FileDialog *Self, STRING Value)
{
   WORD i;

   if (Self->FilterList) { FreeMemory(Self->FilterList); Self->FilterList = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->FilterList, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->FilterList[i] = Value[i];
         Self->FilterList[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Flags
Short:  Optional flags may be set here.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>Optional flags that may be set against a dialog object are as follows:</p>

<table border="1" bordercolordark="#000000" cellspacing="0" cellpadding="2" bordercolorlight="#000000" width="80%" align="center">
<tr><td class="tableheader" bgcolor="#eeeeee">Flag</td><td class="tableheader" bgcolor="#eeeeee">Description</td></tr>
<tr><td width="10%">MULTISELECT</td><td>Enables multi-select mode so that the user may select multiple files and directories.</td></tr>
<tr><td width="10%">STRICTFILTER</td><td>Strictly limits the filter combo box so that the user cannot type in custom filters.</td></tr>
<tr><td width="10%">WAIT</td><td>Normally when a dialog window is displayed, the task's process flow will return immediately to the core program.  This can be problematic if you require a response to the dialog box before continuing with your processing.  To solve this issue, use the WAIT flag to curb the program flow until the dialog box receives a response from the user.</td></tr>
</table>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Icon
Short:  The icon that appears in the window title bar can be defined here.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>A file folder icon is set in the dialog window by default, however you
may change to a different icon image if you wish.  If you are going to refer
to a stock icon, use the file format, "icons:category/name".</p>

###END###

*****************************************************************************/

static ERROR GET_Icon(struct FileDialog *Self, STRING *Value)
{
   if (Self->Icon[0]) {
      *Value = Self->Icon;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Icon(struct FileDialog *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->Icon)-1) AND (Value[i]); i++) Self->Icon[i] = Value[i];
      Self->Icon[i] = 0;
   }
   else Self->Icon[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   OkayText
Short:  Defines the string to use for the okay button.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The text that appears in the okay button of the file dialog may be
redefined here.  By default, the dialog will display the words "Open File"
in the button.</p>

###END###

*****************************************************************************/

static ERROR GET_OkayText(struct FileDialog *Self, STRING *Value)
{
   if (Self->OkayText[0]) {
      *Value = Self->OkayText;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_OkayText(struct FileDialog *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->OkayText)-1) AND (Value[i]); i++) Self->OkayText[i] = Value[i];
      Self->OkayText[i] = 0;
   }
   else Self->OkayText[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Location
Short:  The starting location of the file view is defined here.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>By default the file-view will display the filesystem's root file structure
when it is shown for the first time.  You can switch to a different default directory
on initialisation by setting the Location field.  The file dialog object will
automatically strip-out any file references that trail the location string.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct FileDialog *Self, STRING *Value)
{
   if (Self->Location[0]) {
      *Value = Self->Location;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Location(struct FileDialog *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->Location)-1); i++) Self->Location[i] = Value[i];
      Self->Location[i] = 0;
   }
   else Self->Location[0] = 0;

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   SelectedFiles
Short:  Returns a list of all selected files, assuming a successful open procedure.
Status: Read
Type:   STRING *

###DESCRIPTION###
<p>When using multi-select mode (as defined in the Flags field), you will need
to use the SelectedFiles field when reading a complete list of the files
that the user has selected.  This field will return an array of strings, each
revealing the complete path to a user-selected file.  The array is terminated
with a NULL entry.</p>

<p>The array remains valid up until the next time that you read the
SelectedFiles field.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Static
Short:  Set to TRUE to make the object static.
Status: Read/Write
Type:   OBJECTID

###DESCRIPTION###
<p>By default, the FileDialog object will execute itself and then self-destruct
when a closing tag is received.  If you would rather that the object stays in
the system, set this field to TRUE.  If you do this, the only way to get a
Dialog object to perform is to call the Show action.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Target
Short:  The target for the dialog box window is specified here.
Type:   OBJECTID
Status: Init

###DESCRIPTION###
<p>The window for a dialog box will normally be created on the desktop.  On
occasion it may be useful to have the window appear in a different area, such as
inside another window or screen.  To do this, point the Target field to the
unique ID of the drawable that you want to open the window on.</p>

<p>The target may not be changed after initialisation.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Title
Short:  Defines the window title for the dialog box.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>The window title for the dialog box is specified here as a standard UTF-8
text string.</p>

###END###

*****************************************************************************/

static ERROR GET_Title(struct FileDialog *Self, STRING *Value)
{
   if (Self->Title[0]) {
      *Value = Self->Title;
      return(ERR_Okay);
   }
   else return(ERR_FieldNotSet);
}

static ERROR SET_Title(struct FileDialog *Self, STRING Value)
{
   WORD i;

   if (Value) {
      for (i=0; (i < sizeof(Self->Title)-1) AND (Value[i] >= 0x20); i++) Self->Title[i] = Value[i];
      Self->Title[i] = 0;
   }
   else Self->Title[0] = 0;

   /*** Update the window title ***/

   if (Self->WindowID) {


   }

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Window
Short:  Refers to the ID of the window created by the file dialog object.
Type:   OBJECTID
Status: Read

###DESCRIPTION###
<p>This readable field references the ID of the dialog box's window.  It is
only usable on successful initialisation of a dialog box.  It is recommended
that you avoid tampering with the generated window, but direct access may be
useful for actions such as altering the window position.</p>

###END###

*****************************************************************************/

/*****************************************************************************
** Internal: CreateWindow()
*/

static UBYTE xmlToolbar[] = "\n\
<item text=\"Root\" tip=\"Root Directory\" icon=\"devices/harddisk\"><set object=\"[fview]\" &location=\"\"/></item>\n\
<item text=\"Parent\" tip=\"Parent Folder\" icon=\"arrows/green_nw\"><action object=\"[fview]\" method=\"parentdirectory\"/></item>\n\
<break/>\n\
<item text=\"Cut\" tip=\"Cut\" icon=\"tools/cut\"><action object=\"[fview]\" method=\"cutfiles\"/></item>\n\
<item text=\"Copy\" tip=\"Copy\" icon=\"tools/copy\"><action object=\"[fview]\" method=\"copyfiles\"/></item>\n\
<item text=\"Paste\" tip=\"Paste\" icon=\"tools/paste\"><action object=\"[fview]\" method=\"pastefiles\"/></item>\n\
<break/>\n\
<item text=\"Create Dir\" tip=\"Create Directory\" icon=\"folders/folder_new\"><action object=\"[fview]\" method=\"createdir\"/></item>\n\
<item text=\"Rename\" tip=\"Rename\" icon=\"items/charset\"><action object=\"[fview]\" method=\"renamefile\"/></item>\n\
<item text=\"Delete\" tip=\"Delete\" icon=\"tools/eraser\"><action object=\"[fview]\" method=\"deletefiles\"/></item>\n\
<break/>\n\
<item text=\"Column View\" tip=\"Column View\" icon=\"items/view_detailed\"><set object=\"[fview.view]\" &style=\"column\"/></item>\n\
<item text=\"List View\" tip=\"List View\" icon=\"items/view_text\"><set object=\"[fview.view]\" &style=\"list\"/></item>";

static ERROR CreateWindow(struct FileDialog *Self)
{
   struct mtAddTabObject add;
   STRING filter;
   UBYTE text[60], filtertext[60], filtervalue[60], buffer[200];
   OBJECTPTR script, window, tabfocus, toolbar;
   OBJECTID viewdrawable;
   LONG leftmargin, rightmargin, bottommargin, topmargin, toolbar_height;
   LONG buttonheight, buttonwidth, okaywidth, cancelwidth, flags, path_height;
   WORD i;

   DPrintF("CreateWindow()","[FileDialog:%d]", Self->Head.UniqueID);

   /*** Trim any file names out of the location path ***/

   for (i=0; Self->Location[i]; i++);
   while ((i > 0) AND (Self->Location[i-1] != ':') AND (Self->Location[i-1] != '/') AND (Self->Location[i-1] != '\\')) i--;
   Self->Location[i] = 0;

   Self->Response = RSF_CANCEL;

   if (NewObject(ID_WINDOW, NF_CHILD, &window, &Self->WindowID) IS ERR_Okay) {
      SetFields(window,
         FID_Title|TSTRING,      Self->Title,
         FID_InsideWidth|TLONG,  300,
         FID_InsideHeight|TLONG, 350,
         FID_MinWidth|TLONG,     250,
         FID_MinHeight|TLONG,    150,
         FID_Quit|TLONG,         FALSE,
         FID_Center|TLONG,       TRUE,
         FID_Icon|TSTRING,       Self->Icon,
         TAGEND);

      if (Self->TargetID) SetField(window, FID_Parent, FT_LONG, &Self->TargetID);

      if (Action(AC_Init, window, NULL) IS ERR_Okay) {

         SubscribeActionTags(window, Self, AC_Free, TAGEND);

         GetFields(window, FID_LeftMargin|TLONG,   &leftmargin,
                           FID_TopMargin|TLONG,    &topmargin,
                           FID_BottomMargin|TLONG, &bottommargin,
                           FID_RightMargin|TLONG,  &rightmargin,
                           TAGEND);
      }
      else {
         Action(AC_Free, window, NULL);
         ReleaseObject(window);
         Self->WindowID = NULL;
         return(ObjectError(Self, ERH_Init, ERR_Init));
      }

      ReleaseObject(window);
   }
   else return(ObjectError(Self, ERH_Init, ERR_NewObject));

   /*** Create tab-focus manager ***/

   CreateObject(ID_TABFOCUS, NULL, &tabfocus, NULL,
      FID_Owner|TLONG, Self->WindowID,
      TAGEND);

   /*** Create the toolbar at the top of the window ***/

   if (CreateObject(ID_TOOLBAR, NULL, &toolbar, NULL,
      FID_Owner|TLONG,    Self->WindowID,
      FID_XCoord|TLONG,   leftmargin,
      FID_YCoord|TLONG,   topmargin,
      FID_XOffset|TLONG,  rightmargin,
      TAGEND) IS ERR_Okay) {

      GetField(toolbar, FID_Height, FT_LONG, &toolbar_height);

      acDataChannel(toolbar, Self->Head.UniqueID, DATA_XML, 1, xmlToolbar, StrLength(xmlToolbar) + 1, 1);
      acShow(toolbar);
   }

   /*** Create an input box that will display the current path ***/

   if (CreateObject(ID_INPUT, NULL, &Self->PathBox, NULL,
      FID_Owner|TLONG,    Self->WindowID,
      FID_XCoord|TLONG,   leftmargin,
      FID_YCoord|TLONG,   topmargin + toolbar_height + 4,
      FID_XOffset|TLONG,  rightmargin,
      FID_String|TSTRING, Self->Location,
      FID_Colour|TSTRING,    "#ffffff70",
      FID_Highlight|TSTRING, "#00000050",
      FID_Shadow|TSTRING,    "#00000050",
      TAGEND) IS ERR_Okay) {

      GetField(Self->PathBox, FID_Height, FT_LONG, &path_height);

      SubscribeAction(Self->PathBox, AC_Activate, Self);

      acShow(Self->PathBox);
   }

   /*** Create the open and cancel buttons ***/

   if (CreateObject(ID_BUTTON, NULL, &Self->CancelButton, NULL,
      FID_Owner|TLONG,    Self->WindowID,
      FID_Text|TSTRING,   Self->CancelText,
      FID_XOffset|TLONG,  rightmargin,
      FID_YOffset|TLONG,  bottommargin,
      TAGEND) IS ERR_Okay) {

      GetFields(Self->CancelButton, FID_Height|TLONG, &buttonheight,
                                    FID_Width|TLONG,  &cancelwidth,
                                    TAGEND);

      SubscribeAction(Self->CancelButton, AC_Activate, Self);

      acShow(Self->CancelButton);
   }
   else goto error;

   if (CreateObject(ID_BUTTON, NULL, &Self->OkayButton, NULL,
      FID_Owner|TLONG,    Self->WindowID,
      FID_Text|TSTRING,   Self->OkayText,
      FID_XOffset|TLONG,  rightmargin,
      FID_YOffset|TLONG,  bottommargin + buttonheight + BUTTONGAP,
      TAGEND) IS ERR_Okay) {

      GetField(Self->OkayButton, FID_Width, FT_LONG, &okaywidth);

      SubscribeAction(Self->OkayButton, AC_Activate, Self);

      acShow(Self->OkayButton);
   }
   else goto error;

   if (okaywidth < cancelwidth) {
       SetField(Self->OkayButton, FID_Width, FT_LONG, &cancelwidth);
       buttonwidth = cancelwidth;
   }
   else {
      SetField(Self->CancelButton, FID_Width, FT_LONG, &okaywidth);
      buttonwidth = okaywidth;
   }

   /*** Create the file and filter input boxes ***/

   if (CreateObject(ID_INPUT, NULL, &Self->LocationBox, NULL,
      FID_Owner|TLONG,      Self->WindowID,
      FID_LabelWidth|TLONG, 40,
      FID_XCoord|TLONG,     leftmargin,
      FID_YOffset|TLONG,    bottommargin + buttonheight + BUTTONGAP,
      FID_XOffset|TLONG,    rightmargin + buttonwidth + BUTTONGAP,
      FID_Label|TSTRING,    "File:",
      FID_Height|TLONG,     buttonheight,
      TAGEND) IS ERR_Okay) {

      SubscribeAction(Self->LocationBox, AC_Activate, Self);
      acShow(Self->LocationBox);
   }
   else goto error;

   /*** Determine the text that we are going to display in the filter box ***/

   StrCopy(Self->Filter, filtertext, sizeof(filtertext));

   if ((!Self->Filter[0]) AND (Self->FilterList[0])) {
      filter = Self->FilterList;
      for (i=0; (*filter) AND (*filter != ';') AND (i < sizeof(filtertext)-1); i++) filtertext[i] = *filter++;
      filtertext[i] = 0;
      while ((*filter) AND (*filter != ';')) filter++;
      if (*filter IS ';') filter++;

      for (i=0; (*filter) AND (*filter != ';') AND (i < sizeof(Self->Filter)-1); i++) Self->Filter[i] = *filter++;
      Self->Filter[i] = 0;
   }
   else {
      filter = Self->FilterList;
      while (*filter) {
         for (i=0; (*filter) AND (*filter != ';') AND (i < sizeof(text)-1); i++) text[i] = *filter++;
         text[i] = 0;
         if (*filter IS ';') filter++;

         for (i=0; (*filter) AND (*filter != ';') AND (i < sizeof(filtervalue)-1); i++) filtervalue[i] = *filter++;
         filtervalue[i] = 0;
         if (*filter IS ';') filter++;

         if (StrCompare(filtervalue, Self->Filter, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            StrCopy(text, filtertext, sizeof(filtertext));
            break;
         }
      }
   }

   if (Self->Flags & FDF_STRICTFILTER) flags = CMF_LIMITTOLIST|CMF_AUTOCOMPLETE;
   else flags = CMF_EDIT;

   if (CreateObject(ID_COMBOBOX, NULL, &Self->FilterBox, NULL,
      FID_Owner|TLONG, Self->WindowID,
      FID_LabelWidth|TLONG, 40,
      FID_XCoord|TLONG,     leftmargin,
      FID_YOffset|TLONG,    bottommargin,
      FID_XOffset|TLONG,    rightmargin + buttonwidth + BUTTONGAP,
      FID_Label|TSTRING,    "Filter:",
      FID_Height|TLONG,     buttonheight,
      FID_Flags|TLONG,      flags,
      FID_String|TSTRING,   filtertext,
      TAGEND) IS ERR_Okay) {

      filter = Self->FilterList;
      while (*filter) {
         for (i=0; (*filter) AND (*filter != ';') AND (i < sizeof(text)-1); i++) text[i] = *filter++;
         text[i] = 0;
         if (*filter IS ';') filter++;

         for (i=0; (*filter) AND (*filter != ';') AND (i < sizeof(filtervalue)-1); i++) filtervalue[i] = *filter++;
         filtervalue[i] = 0;
         if (*filter IS ';') filter++;

         StrFormat(buffer, sizeof(buffer), "<item text=\"%s\"/>", text, filtervalue);

         ActionTags(AC_DataChannel, Self->FilterBox, Self->Head.UniqueID, DATA_XML, 1, buffer, StrLength(buffer), 1);
      }

      SubscribeAction(Self->FilterBox, AC_Activate, Self);
      acShow(Self->FilterBox);
   }
   else goto error;

   /*** Create the file view area ***/

   if (CreateObject(ID_RENDER, NULL, NULL, &viewdrawable,
      FID_Name|TSTRING,  "_fdrender",
      FID_Owner|TLONG,   Self->WindowID,
      FID_XCoord|TLONG,  leftmargin,
      FID_YCoord|TLONG,  topmargin + toolbar_height + 4 + path_height + 5,
      FID_XOffset|TLONG, rightmargin,
      FID_YOffset|TLONG, bottommargin + (buttonheight * 2) + (BUTTONGAP * 2),
      TAGEND) != ERR_Okay) goto error;

   acShowID(viewdrawable);

   if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
      FID_Target|TLONG,     viewdrawable,
      FID_Location|TSTRING, "templates:insideborder",
      FID_Input|TUNLISTED,  "1",
      FID_Fill|TUNLISTED,   "1",
      TAGEND) IS ERR_Okay) {
      acActivate(script);
      acFree(script);
   }
   else goto error;

   if (CreateObject(ID_SCROLLBAR, NULL, NULL, NULL,
      FID_Owner|TLONG,       viewdrawable,
      FID_Direction|TSTRING, "vertical",
      FID_XOffset|TTRANS,    "[gl_interface.inputborder_xo]",
      FID_YCoord|TTRANS,     "[gl_interface.inputborder_y]",
      FID_YOffset|TTRANS,    "[gl_interface.inputborder_yo]",
      TAGEND) != ERR_Okay) goto error;

   if (CreateObject(ID_SCROLLBAR, NULL, NULL, NULL,
      FID_Owner|TLONG,       viewdrawable,
      FID_Direction|TSTRING, "horizontal",
      FID_XCoord|TTRANS,     "[gl_interface.inputborder_x]",
      FID_XOffset|TTRANS,    "[gl_interface.inputborder_yo]",
      FID_YOffset|TTRANS,    "[gl_interface.inputborder_yo]",
      TAGEND) != ERR_Okay) goto error;

   if (CreateObject(ID_VIEW, NULL, &Self->View, NULL,
      FID_Owner|TLONG,                 viewdrawable,
      FID_BorderOffset|TTRANS|TSTRING, "[gl_interface.inputborder]",
      FID_Flags|TLONG,                 ((Self->Flags & FDF_MULTISELECT) ? VWF_MULTISELECT : 0),
      TAGEND) IS ERR_Okay) {
      SubscribeField(Self->View, FID_SelectedTag, Self);
   }
   else goto error;

   if (CreateObject(ID_FILEVIEW, NULL, &Self->FileView, NULL,
      FID_Name|TSTRING,     "fview",
      FID_Owner|TLONG,      Self->View->UniqueID,
      FID_View|TLONG,       Self->View->UniqueID,
      FID_Location|TSTRING, Self->Location,
      FID_Filter|TSTRING,   Self->Filter,
      FID_Flags|TLONG,      ((Self->Flags & FDF_NOFILES) ? FVF_NOFILES : 0),
      TAGEND) IS ERR_Okay) {

      SubscribeAction(Self->FileView, AC_Activate, Self);
      SubscribeField(Self->FileView, FID_Location, Self);
   }
   else goto error;

   /*** Define the focussing order for the tab key ***/

   add.ObjectID = Self->PathBox->UniqueID;
   Action(MT_AddTabObject, tabfocus, &add);

   add.ObjectID = viewdrawable;
   Action(MT_AddTabObject, tabfocus, &add);

   add.ObjectID = Self->LocationBox->UniqueID;
   Action(MT_AddTabObject, tabfocus, &add);

   add.ObjectID = Self->FilterBox->UniqueID;
   Action(MT_AddTabObject, tabfocus, &add);

   add.ObjectID = Self->OkayButton->UniqueID;
   Action(MT_AddTabObject, tabfocus, &add);

   add.ObjectID = Self->CancelButton->UniqueID;
   Action(MT_AddTabObject, tabfocus, &add);

   return(ERR_Okay);

error:
   ActionMsg(AC_Free, Self->WindowID, NULL);
   return(ERR_Failed);
}

/*****************************************************************************
** Internal: OpenFile()
*/

static LONG OpenFile(struct FileDialog *Self)
{
   struct XMLTag **taglist;
   LONG *tags;
   OBJECTPTR dialog, xml;
   STRING path, str;
   WORD i, pathlen;
   LONG count, type, size;
   UBYTE buffer[512];

   Self->Response = RSF_CANCEL;

   GetField(Self->FileView, FID_Location, FT_STRING, &path);

   /*** Generate the SelectedFiles array ***/

   if (Self->SelectedFiles) {
      FreeMemory(Self->SelectedFiles);
      Self->SelectedFiles = NULL;
   }

   count = 0;
   if (GetFields(Self->View, FID_SelectedTags|TPTR, &tags,
                             FID_XML|TPTR,          &xml,
                             FID_Tags|TPTR,         &taglist,
                             TAGEND) IS ERR_Okay) {
      /*** Determine how big our memory allocation is going to be ***/

      pathlen = StrLength(path);
      size = sizeof(STRING); /* +Empty array entry at the end */
      for (count=0; tags[count] != -1; count++) {
         ActionTags(MT_GetXMLContent, xml, taglist[tags[count]]->Index, buffer, sizeof(buffer));
         size += sizeof(STRING) + pathlen + StrLength(buffer) + 1;
      }

      DPrintF("OpenFile:","%d items have been selected.", count);

      if (AllocMemory(size, MEM_DATA|MEM_NOCLEAR, &Self->SelectedFiles, NULL) IS ERR_Okay) {
         str = (STRING)(Self->SelectedFiles + count + 1);
         for (i=0; tags[i] != -1; i++) {
            Self->SelectedFiles[i] = str;
            str += StrCopy(path, str, COPY_ALL);
            if (ActionTags(MT_GetXMLContent, xml, taglist[tags[i]]->Index, buffer, sizeof(buffer)) IS ERR_Okay) {
               str += StrCopy(buffer, str, COPY_ALL) + 1;
            }
         }
         Self->SelectedFiles[i] = NULL;
      }
      else ObjectError(Self, ERH_Function, ERR_AllocMemory);
   }
   else DPrintF("OpenFile:","No items have been selected by the user.");

   /*** Determine the location of the primary selected file ***/

   Self->Location[0] = 0;
   i = StrCopy(path, Self->Location, sizeof(Self->Location));

   if ((GetField(Self->LocationBox, FID_String, FT_STRING, &str) IS ERR_Okay) AND (str)) {
      StrCopy(str, Self->Location+i, sizeof(Self->Location));
   }
   else {
      /*** If nothing is typed into the location box, do not continue ***/
      if ((count > 0) AND (Self->Flags & FDF_MULTISELECT));
      else if (!(Self->Flags & FDF_SELECTDIR)) return(FALSE);
   }

   DPrintF("OpenFile()","[FileDialog] %s", Self->Location);

   /*** Check if the file exists ***/

   if (Self->Flags & FDF_CONFIRMEXISTS) {
      /* In confirmation mode, we check to see if the file exists and
      ** if it does, we ask the user if he wants to proceed with
      ** overwriting the file.
      */
      if ((AnalyseLocation(Self->Location, &type) IS ERR_Okay) AND (type IS LOC_FILE)) {
         UBYTE buffer[400];
         LONG response;

         StrFormat(buffer, sizeof(buffer), "You are about to overwrite the file '%s'.  Are you sure that you wish to continue?", Self->Location);

         if (CreateObject(ID_DIALOG, NF_CHILD, &dialog, NULL,
            FID_Target|TLONG,    Self->TargetID,
            FID_Flags|TLONG,     DF_WAIT,
            FID_Static|TLONG,    TRUE,
            FID_Buttons|TSTRING, "yes;no",
            FID_Title|TSTRING,   "Confirmation Required",
            FID_Image|TSTRING,   "icons:items/warning(48)",
            FID_String|TSTRING,  buffer,
            TAGEND) IS ERR_Okay) {

            Action(AC_Show, dialog, NULL);

            GetField(dialog, FID_Response, FT_LONG, &response);

            Action(AC_Free, dialog, NULL);

            if (response != RSF_YES) return(FALSE);
         }
      }
   }
   else if ((!(Self->Flags & FDF_NOFILECHECK)) AND (!(Self->Flags & FDF_MULTISELECT))) {
      if (AnalyseLocation(Self->Location, &type) != ERR_Okay) {
         DPrintF("OpenFile:","[FileDialog] Failed to analyse the location.");
         return(FALSE);
      }

      if (Self->Flags & FDF_SELECTDIR) {
         if ((type != LOC_FILE) AND (type != LOC_DIRECTORY) AND (type != LOC_ASSIGN)) {
            DPrintF("OpenFile:","[FileDialog] Location is neither directory nor file.");
            return(FALSE);
         }
      }
      else if (type != LOC_FILE) {
         DPrintF("OpenFile:","[FileDialog] Location is not a file.");
         return(FALSE);
      }
   }

   Self->Response = RSF_OKAY;

   /*** Auto-Save ***/

   if (Self->AutoSaveID) {
      OBJECTPTR file, autosave;

      if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
         FID_Location|TSTRING, Self->Location,
         FID_Flags|TLONG,      FL_WRITE|FL_NEW,
         TAGEND) IS ERR_Okay) {

         if (AccessObject(Self->AutoSaveID, 3000, &autosave) IS ERR_Okay) {
            ActionTags(AC_SaveToObject, autosave, file->UniqueID, 0);
            ReleaseObject(autosave);
         }
         else {
            Action(AC_Free, file, NULL);
            return(FALSE);
         }

         Action(AC_Free, file, NULL);
      }
      else return(FALSE);
   }

   /*** Run the action script ***/

   if (GET_ActionScript(Self, &Self->Script) IS ERR_Okay) {
      if (!Self->ScriptObject) {
         CreateObject(ID_SCRIPT, NF_CHILD, &Self->ScriptObject, NULL,
            FID_Location|TSTRING, Self->Script,
            TAGEND);
      }

      SetField(Self->ScriptObject, FID_Target, FT_LONG, &Self->WindowID);
      SetUnlistedField(Self->ScriptObject, "location", Self->Location);
      Action(AC_Activate, Self->ScriptObject, NULL);
   }

   /*** Activate child objects ***/

   if ((count = TotalChildren(Self->Head.UniqueID))) {
      struct ChildEntry list[count];

      if (ListChildren(Self->Head.UniqueID, list, &count) IS ERR_Okay) {
         for (i=0; i < count; i++) {
            ActionMsg(AC_Activate, list[i].ObjectID, NULL);
         }
      }
   }

   /*** Provide Activate notification ***/

   NotifySubscribers(Self, AC_Activate, NULL, NULL);

   return(TRUE);
}
